<<<<<<< HEAD:flakytests/cylc-reset/02-output-1.t
#!/bin/bash
# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
=======
#!/usr/bin/env bash
# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
>>>>>>> upstream/8.0_b2:tests/functional/job-submission/00-user.t
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#-------------------------------------------------------------------------------
# Test user-defined job runner handlers can be used.
. "$(dirname "$0")/test_header"
set_test_number 2
<<<<<<< HEAD:flakytests/cylc-reset/02-output-1.t
=======

create_test_global_config "" "
[platforms]
  [[testme]]
    hosts = localhost
    job runner = my_background
    install target = localhost
  [[testme2]]
    hosts = localhost
    job runner = my_background2
    install target = localhost
"

>>>>>>> upstream/8.0_b2:tests/functional/job-submission/00-user.t
reftest
exit
