[scheduler]
   UTC mode = True
[scheduling]
    initial cycle point = 20100808T00
    final cycle point = 20100809T00
    [[graph]]
        T00 = "foo[-P1D]:restart1 => foo"
[runtime]
    [[foo]]
        script = """
<<<<<<< HEAD:tests/functional/graphing/07-simple-messaging/suite.rc
wait
=======
cylc__job__wait_cylc_message_started
>>>>>>> upstream/8.0_b2:tests/functional/pre-initial/07-simple-messaging/flow.cylc
cylc message "restart files ready"
"""
        [[[outputs]]]
            restart1 = "restart files ready"
