#!jinja2
<<<<<<< HEAD:flakytests/execution-time-limit/04-poll/suite.rc
[cylc]
=======
[scheduler]
>>>>>>> upstream/8.0_b2:tests/flakyfunctional/execution-time-limit/04-poll/flow.cylc
   [[events]]
       abort if any task fails = True
       abort on inactivity = True
       inactivity = PT2M

[scheduling]
    [[graph]]
        R1 = foo

[runtime]
    [[foo]]
        init-script = cylc__job__disable_fail_signals ERR EXIT
        script = """
cylc__job__wait_cylc_message_started
if [[ "${CYLC_TASK_SUBMIT_NUMBER}" == '1' ]]; then
    # Will be killed after PT5S
    sleep 40
else
    # Fake success
    cat >>"$0.status" <<__STATUS__
CYLC_JOB_EXIT=SUCCEEDED
CYLC_JOB_EXIT_TIME=$(date -u '+%FT%H:%M:%SZ')
__STATUS__
fi
exit 1
"""
        [[[job]]]
            execution time limit = PT10S
            execution retry delays = PT0S
