#!/usr/bin/env bash
# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#-------------------------------------------------------------------------------
# Test ICP cutoff bug.
. "$(dirname "$0")/test_header"
set_test_number 1

init_workflow "${TEST_NAME_BASE}" <<'__WORKFLOW__'
[scheduler]
    cycle point time zone = Z
    [[events]]
        abort on stall timeout = True
        stall timeout = PT0S
[scheduling]
    initial cycle point = 20171101T0000Z
    [[graph]]
        R1 = foo
        R1/T00 = foo[^] => qux
        R1/T06 = foo[^] => qux
[runtime]
    [[foo, qux]]
        script = true
__WORKFLOW__
run_ok "${TEST_NAME_BASE}" cylc play "${WORKFLOW_NAME}" --no-detach

purge
exit
