[meta]
    title = content reload test
    description = """two tasks: the second will fail, causing the test to
<<<<<<< HEAD:tests/functional/workflow-host-self-id/02-content/suite.rc
fail, unless the first reloads the suite definition after modifying it."""
=======
fail, unless the first reloads the workflow definition after modifying it."""
>>>>>>> upstream/8.0_b2:tests/functional/reload/02-content/flow.cylc

[scheduling]
    [[graph]]
        R1 = "reloader => foo"

[runtime]
    [[reloader]]
        script = """
# change the value of $FALSE to "true" in foo's environment:
perl -pi -e 's/(FALSE = )false( # marker)/\1true\2/' $CYLC_WORKFLOW_RUN_DIR/flow.cylc
# reload
cylc reload $CYLC_WORKFLOW_NAME
"""
    [[foo]]
        script = "$FALSE"
        [[[environment]]]
            FALSE = false # marker
