[meta]
    title = final cycle reload test
    description = """change final cycle."""

[scheduler]
   UTC mode = True

[scheduling]
    initial cycle point = 20100101T00
    final cycle point = 20100102T00 # marker
    [[graph]]
        R1 = "reloader => a"
        PT6H = "a[-PT6H] => a"
[runtime]
    [[reloader]]
        script = """
# change the final cycle:
perl -pi -e 's/(final cycle point = )20100102T00( # marker)/\1 20100101T12\2/' $CYLC_WORKFLOW_RUN_DIR/flow.cylc
# reload
<<<<<<< HEAD:tests/functional/workflow-host-self-id/07-final-cycle/suite.rc
cylc reload -f $CYLC_SUITE_NAME
cylc__job__poll_grep_suite_log -F 'Reload completed'
=======
cylc reload $CYLC_WORKFLOW_NAME
cylc__job__poll_grep_workflow_log -F 'Reload completed'
>>>>>>> upstream/8.0_b2:tests/functional/reload/07-final-cycle/flow.cylc
"""
    [[a]]
        script = true
