<<<<<<< HEAD:etc/bin/stop-group-demo.sh
#!/bin/bash

# THIS FILE IS PART OF THE CYLC SUITE ENGINE.
=======
#!/usr/bin/env python3
# THIS FILE IS PART OF THE CYLC WORKFLOW ENGINE.
>>>>>>> upstream/8.0_b2:tests/functional/jinja2/08-local-lib-python/Jinja2Filters/qualify.py
# Copyright (C) NIWA & British Crown (Met Office) & Contributors.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

<<<<<<< HEAD:etc/bin/stop-group-demo.sh
for S in $(cylc scan -n groups | awk '{print $1}'); do
  echo "$S"
  cylc stop "$S"
done
=======

import local_lookup


def qualify(arg):
    return local_lookup.lookup(arg)
>>>>>>> upstream/8.0_b2:tests/functional/jinja2/08-local-lib-python/Jinja2Filters/qualify.py
